import json
import os
from pathlib import Path
import tempfile
import unittest

from chronix2grid.config import DispatchConfigManager


class TestConfigManager(unittest.TestCase):
    def setUp(self) -> None:
        self.root_directory = tempfile.mkdtemp()
        os.makedirs(os.path.join(self.root_directory, 'input1'))
        os.makedirs(os.path.join(self.root_directory, 'input2'))
        os.makedirs(os.path.join(self.root_directory, 'output'))
        config_file1 = Path(os.path.join(self.root_directory, 'input1', 'config1.json'))
        config_file1.touch()
        config_file2 = Path(os.path.join(self.root_directory, 'input2', 'config2.json'))
        config_file2.touch()
        self.config_manager_multi = DispatchConfigManager(
            name='TestConfig',
            root_directory=self.root_directory,
            input_directories=dict(first_dir='input1', second_dir='input2'),
            output_directory='output',
            required_input_files=dict(first_dir=['config1.json'],
                                      second_dir=['config2.json'])
        )
        os.makedirs(os.path.join(self.root_directory, 'input'))
        config_file = Path(os.path.join(self.root_directory, 'input', 'config.json'))
        config_file.touch()
        self.config_manager_single = DispatchConfigManager(
            name='TestConfig',
            root_directory=self.root_directory,
            input_directories='input',
            output_directory='output',
            required_input_files=['config.json']
        )

    def test_error_message(self):
        input_directory_abs_path = os.path.join(
            self.root_directory, self.config_manager_single.input_directories)
        output_directory_abs_path = os.path.join(
            self.root_directory, self.config_manager_single.output_directory)
        expected_error_message_single = (
            f"\nTestConfig process requires the following configuration "
            f"(directories and files must exist): \n"
            f"  - Root directory is {self.root_directory}\n"
            f"  - Output directory is {output_directory_abs_path}\n"
            f"  - Input directory is {input_directory_abs_path}\n"
            f"    - Expected input files are:\n"
            f"      - config.json"
        )

        expected_error_message_multi = (
            f"\nTestConfig process requires the following configuration "
            f"(directories and files must exist): \n"
            f"  - Root directory is {self.root_directory}\n"
            f"  - Output directory is {output_directory_abs_path}\n"
            f"  - Input directories are:"
            f"\n    - {os.path.join(self.root_directory, 'input1')}"
            f"\n      - with expected input files:"
            f"\n        - config1.json"
            f"\n    - {os.path.join(self.root_directory, 'input2')}"
            f"\n      - with expected input files:"
            f"\n        - config2.json"
        )
        self.assertEqual(
            self.config_manager_single.error_message(),
            expected_error_message_single
        )
        self.assertEqual(
            self.config_manager_multi.error_message(),
            expected_error_message_multi
        )

    def test_input_mode(self):
        self.assertTrue(self.config_manager_single.is_single_input_dir())
        self.assertFalse(self.config_manager_multi.is_single_input_dir())

    def test_raise_input_directory_does_not_exists(self):
        self.config_manager_single.input_directories = 'input_with_typo'
        with self.assertRaises(FileNotFoundError) as cm:
            self.config_manager_single.validate_input()
        self.assertEqual(str(cm.exception), self.config_manager_single.error_message())

    def test_validate_output_directory_fails(self):
        self.config_manager_single.output_directory = 'output_with_typo'
        self.assertFalse(self.config_manager_single.validate_output())

    def test_has_required_inputs(self):
        self.assertTrue(self.config_manager_single.validate_input())
        Path(os.path.join(self.root_directory, 'input', 'randomfile.json')).touch()
        self.assertTrue(self.config_manager_single.validate_input())

        self.config_manager_single.required_input_files.append('missingconfig.json')
        self.assertFalse(self.config_manager_single.validate_input())

    def test_check_configuration(self):
        self.config_manager_single.output_directory = 'output_with_typo'
        with self.assertRaises(FileNotFoundError) as cm:
            self.config_manager_single.validate_configuration()
        self.assertEqual(str(cm.exception), self.config_manager_single.error_message())

        self.config_manager_single.output_directory = 'output'
        self.config_manager_single.input_directories = 'input_with_typo'
        with self.assertRaises(FileNotFoundError) as cm:
            self.config_manager_single.validate_configuration()
        self.assertEqual(str(cm.exception), self.config_manager_single.error_message())

        self.config_manager_single.input_directories = 'input'
        self.config_manager_single.required_input_files = ['config_with_typo.json']
        with self.assertRaises(FileNotFoundError) as cm:
            self.config_manager_single.validate_configuration()
        self.assertEqual(str(cm.exception), self.config_manager_single.error_message())

    def test_read_config(self):
        self.config_manager_multi.input_directories.update(params='input1')
        self.config_manager_multi.required_input_files.update(params=['params_opf.json'])
        params_opf_filepath = os.path.join(self.root_directory, 'input1', 'params_opf.json')
        Path(params_opf_filepath).touch()
        with open(params_opf_filepath, 'w') as params_opf_file:
            json.dump(dict(test=3), params_opf_file)
        json_dict = self.config_manager_multi.read_configuration()
        self.assertEqual(json_dict["test"], 3)









# Native python libraries
import os
import json

# Other Python libraries
import pandas as pd
import numpy as np
from datetime import timedelta


# Libraries developed for this module
from .consumption import generate_load as gen_loads
from .renewable import generate_solar_wind as gen_enr
from .thermal import generate_dispatch as gen_dispatch
from .dispatch import utils as du
from .dispatch import EconomicDispatch as ec
from . import generation_utils as gu
from ..config import DispatchConfigManager, LoadsConfigManager, ResConfigManager


def time_parameters(weeks, start_date):
    result = dict()
    start_date = pd.to_datetime(start_date, format='%Y-%m-%d')
    result['weeks'] = weeks
    result['start_date'] = start_date
    result['year'] = start_date.year
    return result


def updated_time_parameters_with_timestep(time_parameters, timestep):
    time_parameters['end_date'] = time_parameters['start_date'] + timedelta(
        days=7 * int(time_parameters['weeks'])) - timedelta(minutes=timestep)
    time_parameters['T'] = int(
        pd.Timedelta(
            time_parameters['end_date'] - time_parameters['start_date']
        ).total_seconds() // 60
    )
    return time_parameters


# Call generation scripts n_scenario times with dedicated random seeds
def main(case, n_scenarios, input_folder, output_folder, time_params, mode='LRTK'):
    """
    Main function for chronics generation. It works with three steps: load generation, renewable generation (solar and wind) and then dispatch computation to get the whole energy mix

    Parameters
    ----------
    case (str): name of case to study (must be a folder within input_folder)
    year (int): year of generation
    n_scenarios (int): number of desired scenarios to generate for the same timescale
    params (dict): parameters of generation, as returned by function chronix2grid.generation.generate_chronics.read_configuration
    input_folder (str): path of folder containing inputs
    output_folder (str): path where outputs will be written (intermediate folder case/year/scenario will be used)
    prods_charac (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    loads_charac (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    lines (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    solar_pattern (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    load_weekly_pattern (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    mode (str): options to launch certain parts of the generation process : L load R renewable T thermal

l
    Returns
    -------

    """

    print('=====================================================================================================================================')
    print('============================================== CHRONICS GENERATION ==================================================================')
    print('=====================================================================================================================================')

    ## Random seeds:  Make sure the seeds are the same, whether computation is parallel or sequential
    seeds = [np.random.randint(low=0, high=2 ** 31) for _ in range(n_scenarios)]

    ## Folder settings
    year = time_params['year']

    dispatch_input_folder, dispatch_input_folder_case, dispatch_output_folder = gu.make_generation_input_output_directories(input_folder, case, year, output_folder)
    load_config_manager = LoadsConfigManager(
        name="Loads Generation",
        root_directory=input_folder,
        input_directories=dict(case=case, patterns='patterns'),
        required_input_files=dict(case=['loads_charac.csv', 'params.json'],
                                  patterns=['load_weekly_pattern.csv']),
        output_directory=dispatch_input_folder
    )
    load_config_manager.validate_configuration()

    params, loads_charac, load_weekly_pattern = load_config_manager.read_configuration()

    res_config_manager = ResConfigManager(
        name="Renewables Generation",
        root_directory=input_folder,
        input_directories=dict(case=case, patterns='patterns'),
        required_input_files=dict(case=['prods_charac.csv', 'params.json'],
                                  patterns=['solar_pattern.npy']),
        output_directory=dispatch_input_folder
    )

    params, prods_charac, solar_pattern = res_config_manager.read_configuration()

    params.update(time_params)
    params = updated_time_parameters_with_timestep(params, params['dt'])

    dispath_config_manager = DispatchConfigManager(
        name="Dispatch",
        root_directory=os.path.join(input_folder, case),
        output_directory=dispatch_output_folder,
        input_directories=dict(params='.', year=os.path.join('dispatch', str(year))),
        required_input_files=dict(params=['params_opf.json'], year=[])
    )
    dispath_config_manager.validate_configuration()
    params_opf = dispath_config_manager.read_configuration()
    dispatcher = ec.init_dispatcher_from_config(params_opf["grid_path"], input_folder)

    ## Launch proper scenarios generation
    for i, seed in enumerate(seeds):
        scenario_name = f'Scenario_{i}'
        scenario_dispatch_input_folder = os.path.join(dispatch_input_folder, scenario_name)

        print("================ Generating "+scenario_name+" ================")
        if 'L' in mode:
            load, load_forecasted = gen_loads.main(scenario_dispatch_input_folder, seed, params, loads_charac, load_weekly_pattern, write_results = True)

        if 'R' in mode:
            prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted = gen_enr.main(scenario_dispatch_input_folder, seed, params, prods_charac, solar_pattern, write_results = True)
        if 'T' in mode:
            input_scenario_folder, output_scneario_folder = du.make_scenario_input_output_directories(
                dispatch_input_folder, dispatch_output_folder, scenario_name)
            prods = pd.concat([prod_solar, prod_wind], axis=1)
            res_names = dict(wind=prod_wind.columns, solar=prod_solar.columns)
            dispatcher.chronix_scenario = ec.ChroniXScenario(load, prods, res_names,
                                                             scenario_name)

            dispatch_results = gen_dispatch.main(dispatcher, input_scenario_folder,
                                                 output_scneario_folder,
                                                 seed, params_opf)
        print('\n')
    return params, loads_charac, prods_charac




from abc import ABC, abstractmethod
import json
import os

import numpy as np
import pandas as pd

from chronix2grid.generation.dispatch import utils as du


class ConfigManager(ABC):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        self.name = name
        self.root_directory = root_directory
        self.input_directories = input_directories
        self.output_directory = output_directory
        self.required_input_files = required_input_files if required_input_files is not None else []

    def is_single_input_dir(self):
        if isinstance(self.input_directories, str):
            return True
        if isinstance(self.input_directories, dict):
            return False
        raise RuntimeError("input_directories must be either a string or a dictionnary")

    def validate_input(self):
        if self.is_single_input_dir():
            directory_to_check = os.path.join(self.root_directory, self.input_directories)
            try:
                files_to_check = os.listdir(directory_to_check)
                for config_file in self.required_input_files:
                    if config_file not in files_to_check:
                        return False
            except FileNotFoundError:
                raise FileNotFoundError(self.error_message())
        else:
            for input_name, input_path in self.input_directories.items():
                input_directory_abs_path = os.path.join(self.root_directory, input_path)
                try:
                    files_to_check = os.listdir(input_directory_abs_path)
                    for config_file in self.required_input_files[input_name]:
                        if config_file not in files_to_check:
                            return False
                except FileNotFoundError:
                    raise FileNotFoundError(self.error_message())
        return True

    def validate_output(self):
        return os.path.exists(os.path.join(self.root_directory, self.output_directory))

    def error_message(self):
        output_directory_abs_path = os.path.join(
            self.root_directory, self.output_directory)
        error_message_header = (
            f"\n{self.name} process requires the following configuration "
            f"(directories and files must exist): \n"
            f"  - Root directory is {self.root_directory}\n"
            f"  - Output directory is {output_directory_abs_path}\n"
        )
        if self.is_single_input_dir():
            input_directory_abs_path = os.path.join(
                self.root_directory, self.input_directories)
            formatted_required_input_files = ""
            for filename in sorted(self.required_input_files):
                formatted_required_input_files += "\n      - " + filename

            error_msg_body = (
                f"  - Input directory is {input_directory_abs_path}\n"
                f"    - Expected input files are:" + formatted_required_input_files
            )
        else:
            formatted_required_directories = ""
            for input_name, input_path in self.input_directories.items():
                input_directory_abs_path = os.path.join(self.root_directory, input_path)
                formatted_required_directories += (
                        "\n    - " + input_directory_abs_path +
                        "\n      - with expected input files:"
                )
                for required_file in self.required_input_files[input_name]:
                    formatted_required_directories += "\n        - "+required_file

            error_msg_body = (
                    f"  - Input directories are:" + formatted_required_directories
            )
        return error_message_header + error_msg_body

    def validate_configuration(self):
        if not self.validate_output() or not self.validate_input():
            raise FileNotFoundError(self.error_message())
        return True

    @abstractmethod
    def read_configuration(self):
        pass


class LoadsConfigManager(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(LoadsConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        json1_file = open(os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params.json'))
        json1_str = json1_file.read()
        params = json.loads(json1_str)
        for key, value in params.items():
            try:
                params[key] = float(value)
            except ValueError:
                params[key] = pd.to_datetime(value, format='%Y-%m-%d')

        # Nt_inter = int(params['T'] // params['dt'] + 1)
        try:
            loads_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'loads_charac.csv'),
                sep=',')
            names = loads_charac['name']  # to generate error if separator is wrong

        except:
            loads_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'loads_charac.csv'),
                sep=';')


        load_weekly_pattern = pd.read_csv(
            os.path.join(self.root_directory, self.input_directories['patterns'],
                         'load_weekly_pattern.csv'))

        return params, loads_charac, load_weekly_pattern


class ResConfigManager(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(ResConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        json1_file = open(os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params.json'))
        json1_str = json1_file.read()
        params = json.loads(json1_str)
        for key, value in params.items():
            try:
                params[key] = float(value)
            except ValueError:
                params[key] = pd.to_datetime(value, format='%Y-%m-%d')

        # Nt_inter = int(params['T'] // params['dt'] + 1)
        try:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=',')
            names = prods_charac['name']  # to generate error if separator is wrong

        except:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=';')

        solar_pattern = np.load(
            os.path.join(self.root_directory, self.input_directories['patterns'],
                         'solar_pattern.npy'))

        return params, prods_charac, solar_pattern


class DispatchConfigManager(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(DispatchConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        self.validate_configuration()
        params_filepath = os.path.join(
            self.root_directory,
            self.input_directories['params'],
            'params_opf.json')
        with open(params_filepath, 'r') as opf_param_json:
            params_opf = json.load(opf_param_json)
        return params_opf
